#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""B-ASIC Scheduler-gui Graphics Axis Item Module.

Contains the scheduler-gui GraphicsAxisItem class for drawing and maintain the axis in a graph.
"""
from operator import contains
import os
import sys
from typing     import Any, Optional
from pprint     import pprint
from typing     import Any, AnyStr, Generic, Protocol, TypeVar, Union, Optional, overload, Final, final, Dict
# from typing_extensions import Self, Final, Literal, LiteralString, TypeAlias, final
import numpy    as np
from copy       import deepcopy
from math       import cos, sin, pi

import qtpy
from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets

# QGraphics and QPainter imports
from qtpy.QtCore    import (
    Qt, QObject, QRect, QRectF, QPoint, QSize, QSizeF, QByteArray, qAbs)
from qtpy.QtGui     import (
    QPaintEvent, QPainter, QPainterPath, QColor, QBrush, QPen, QFont, QPolygon, QIcon, QPixmap,
    QLinearGradient, QTransform, QPolygonF)
from qtpy.QtWidgets import (
    QGraphicsView, QGraphicsScene, QGraphicsWidget,
    QGraphicsLayout, QGraphicsLinearLayout, QGraphicsGridLayout, QGraphicsLayoutItem, QGraphicsAnchorLayout,
    QGraphicsItem, QGraphicsItemGroup, QGraphicsPathItem, QGraphicsLineItem, QGraphicsTextItem, QGraphicsRectItem,
    QStyleOptionGraphicsItem, QWidget, QGraphicsObject, QGraphicsSimpleTextItem, QGraphicsPolygonItem)
from qtpy.QtCore    import (
    QPoint, QPointF)

# B-ASIC
import logger
from graphics_graph_event   import GraphicsGraphEvent



class GraphicsAxisItem(QGraphicsItemGroup):
    """A class to represent axis in a graph."""
    _scale:         float = 1.0
    """Static, changed from MainWindow."""
    _width_padding: float
    _width:         float
    _height:        float
    _dy_height:     float
    _x_indent:      float
    _axis:          Dict[str, Union[QGraphicsItemGroup, QGraphicsLineItem]]

    def __init__(self, width: float, height: float, x_indent: float  = 0.2, parent: Optional[QGraphicsItem] = None):
        """Constructs a GraphicsAxisItem. 'parent' is passed to QGraphicsItemGroup's constructor."""
        super().__init__(parent)

        self._dy_height = 5/self._scale
        self._axis = {}
        self._width_padding = 0.6
        self._width = width + self._width_padding
        self._height = height
        self._x_indent = x_indent

        self._make_axis()


    def clear(self) -> None:
        """Sets all children's parent to 'None' and delete the axis."""
        keys = list(self._axis.keys())
        for key in keys:
            print(f'clear() key: {key}')
            self._axis[key].setParentItem(None)
            del self._axis[key]
    

    @property
    def width(self) -> float:
        """Get or set the current x-axis width. Setting the width to a new
        value will update the axis automatically."""
        return self._width
    @width.setter
    def width(self, width: float) -> None:
        if self._width != width + self._width_padding:
            self._width = width + self._width_padding
            self.update_axis()

    @property
    def height(self) -> float:
        """Get or set the current y-axis height. Setting the height to a new
        value will update the axis automatically."""
        return self._height
    @height.setter
    def height(self, height: float) -> None:
        if self._height != height:
            self._height = height
            self.update_axis()
    
    @property
    def x_indent(self) -> float:
        """Get or set the current x-axis indent. Setting the indent to a new
        value will update the axis automatically."""
        return self._x_indent
    @x_indent.setter
    def x_indent(self, x_indent: float) -> None:
        if self._x_indent != x_indent:
            self._x_indent = x_indent
            self.update_axis()


    def update_axis(self, width: Optional[float] = None, height: Optional[float] = None, x_indent: Optional[float] = None) -> None:
        """Updates the current axis with the new 'width', 'height' and 'x_indent'. If any of the
        parameters is omitted, the stored value will be used."""
        if width is not None:    self._width = width + self._width_padding
        if height is not None:   self._height = height
        if x_indent is not None: self._x_indent = x_indent
        self.clear()                    # make sure the group is empty
        self._make_axis()               # make new axis


    def _make_axis(self) -> None:
        """Makes new axis out of the stored attributes."""
        ## define pencils
        pen = QPen()
        pen.setWidthF(2/self._scale)
        pen.setJoinStyle(Qt.MiterJoin)
        ledger_pen = QPen(Qt.lightGray)
        ledger_pen.setWidthF(0)         # 0 = cosmetic pen 1px width
        
        ## x-axis
        self._axis['x'] = QGraphicsItemGroup()
        line = QGraphicsLineItem(0, 0, self.width, 0)
        line.setPen(pen)
        self._axis['x'].addToGroup(line)
        # x-axis arrow
        arrow_size = 8/self._scale
        p0 = QPointF(0, sin(pi/6) * arrow_size)
        p1 = QPointF(arrow_size, 0)
        p2 = QPointF(0, -sin(pi/6) * arrow_size)
        polygon = QPolygonF([p0, p1, p2])
        arrow = QGraphicsPolygonItem(polygon)
        arrow.setPen(pen)
        arrow.setBrush(QBrush(Qt.SolidPattern))
        arrow.setPos(self.width, 0)
        self._axis['x'].addToGroup(arrow)
        # x-axis scale
        x_scale = []
        x_scale_labels = []
        x_ledger = []
        self._axis['x_ledger'] = QGraphicsItemGroup()
        for i in range(int(self.width) + 1):
            x_pos = QPointF(self.x_indent + i, 0)
            # vertical x-scale
            x_scale.append(QGraphicsLineItem(0, 0, 0, 0.05))
            x_scale[i].setPen(pen)
            x_scale[i].setPos(x_pos)
            self._axis['x'].addToGroup(x_scale[i])
            # numbers
            x_scale_labels.append(QGraphicsSimpleTextItem(str(i)))
            x_scale_labels[i].setScale(x_scale_labels[i].scale() / self._scale)
            center = x_pos - self.mapFromItem(x_scale_labels[i], x_scale_labels[i].boundingRect().center())
            x_scale_labels[i].setPos(center + QPointF(0, 0.2))
            self._axis['x'].addToGroup(x_scale_labels[i])
            # vertical x-ledger
            x_ledger.append(QGraphicsLineItem(0, 0, 0, self.height))
            if i == int(self.width):                                       # last line is special
                ledger_pen.setWidthF(2/self._scale)
                ledger_pen.setStyle(Qt.DashLine)
                ledger_pen.setColor(Qt.black)
            x_ledger[i].setPen(ledger_pen)
            x_ledger[i].setPos(x_pos)
            self._axis['x_ledger'].addToGroup(x_ledger[i])
        # x-axis label
        label = QGraphicsSimpleTextItem('time')
        label.setScale(label.scale() / self._scale)
        center = self.mapFromItem(arrow, arrow.boundingRect().center())     # =center of arrow
        center -= self.mapFromItem(label, label.boundingRect().center())    # -center of label
        label.setPos(center + QPointF(0, 0.2))                              # move down under arrow
        self._axis['x'].addToGroup(label)

        # y-axis
        self._axis['y'] = QGraphicsLineItem(0, 0, 0, self.height + self._dy_height)
        self._axis['y'].setPen(pen)
        
        # put it all together
        self._axis['x_ledger'].setPos(0, self._dy_height)
        self.addToGroup(self._axis['x_ledger'])
        self._axis['x'].setPos(0, self.height + self._dy_height)
        self.addToGroup(self._axis['x'])
        self.addToGroup(self._axis['y'])
        