#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""B-ASIC Scheduler-gui Graphics Graph Item Module.

Contains the scheduler-gui GraphicsGraphItem class for drawing and
maintain a component in a graph.
"""
import os
import sys
from typing     import Any, Optional
from pprint     import pprint
from typing     import Any, AnyStr, Generic, Protocol, TypeVar, Union, Optional, overload, Final, final, List
# from typing_extensions import Self, Final, Literal, LiteralString, TypeAlias, final
import numpy    as np
from copy       import deepcopy
from itertools  import combinations

import qtpy
from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets

# QGraphics and QPainter imports
from qtpy.QtCore    import (
    Qt, QObject, QRect, QRectF, QPoint, QSize, QSizeF, QByteArray, Slot)
from qtpy.QtGui     import (
    QPaintEvent, QPainter, QPainterPath, QColor, QBrush, QPen, QFont, QPolygon, QIcon, QPixmap,
    QLinearGradient, QTransform)
from qtpy.QtWidgets import (
    QApplication,
    QGraphicsView, QGraphicsScene, QGraphicsWidget,
    QGraphicsLayout, QGraphicsLinearLayout, QGraphicsGridLayout, QGraphicsLayoutItem, QGraphicsAnchorLayout,
    QGraphicsItem, QGraphicsItemGroup, QGraphicsPathItem, QGraphicsLineItem, QGraphicsRectItem,
    QStyleOptionGraphicsItem, QWidget, QGraphicsObject,
    QGraphicsSceneContextMenuEvent, QGraphicsSceneHoverEvent)
from qtpy.QtCore    import (
    QPoint, QPointF, QEvent)

# B-ASIC
import logger
from b_asic.schedule            import Schedule
from b_asic.graph_component     import GraphComponent
from b_asic.special_operations  import Input, Output
from graphics_component_item    import GraphicsComponentItem
from graphics_axis_item         import GraphicsAxisItem
from graphics_graph_event       import GraphicsGraphEvent


class GraphicsGraphItem(QGraphicsItemGroup, GraphicsGraphEvent):
    """A class to represent a graph in a QGraphicsScene. This class is a
    subclass of QGraphicsItemGroup and contains the objects, axes from
    GraphicsAxisItem, as well as components from GraphicsComponentItem. It
    also inherits from GraphicsGraphEvent, which acts as a filter for events
    to GraphicsComponentItem objects."""
    _schedule:          Schedule
    _axis:              GraphicsAxisItem
    _components:        List[GraphicsComponentItem]
    _components_height: float
    _x_axis_indent:     float

    
    def __init__(self, schedule: Schedule, parent: Optional[QGraphicsItem] = None):
        """Constructs a GraphicsGraphItem. 'parent' is passed to QGraphicsItemGroup's constructor."""
        super().__init__(parent)
        self._schedule = deepcopy(schedule)
        self._axis = None
        self._components = []
        self._components_height = 0.0
        self._x_axis_indent = 0.2

        self._make_graph()
        
    def clear(self) -> None:
        """Sets all children's parent to 'None' and delete the children objects."""
        for item in self.childItems():
            item.setParentItem(None)
            del item
    

    def is_valid_pos(self, pos: float) -> bool:
        """Takes in a component position and returns true if the component's new position is valid, false otherwise."""
        # TODO: implement
        # item = self.scene().mouseGrabberItem()
        if pos < 0:
            return False
        return True

        
    @property
    def schedule(self) -> Schedule:
        return self._schedule
    
    @property
    def axis(self) -> GraphicsAxisItem:
        return self._axis
    
    @property
    def items(self) -> list[GraphicsComponentItem]:
        return self._components.childItems()
    
    def _make_graph(self) -> None:
        """Makes a new graph out of the stored attributes."""
        # build components
        spacing = 0.2
        # print('Start times:')
        for op_id, op_start_time in self.schedule.start_times.items():
            op = self.schedule.sfg.find_by_id(op_id)
            
            if not isinstance(op, (Input, Output)):
                self._components_height += spacing
                component = GraphicsComponentItem(op_id, op.latency_offsets, op.execution_time)
                component.setPos(self._x_axis_indent + op_start_time, self._components_height)
                self._components.append(component)
                self._components_height += component.height
        self._components_height += spacing

        # build axis
        schedule_time = self.schedule.schedule_time
        self._axis = GraphicsAxisItem(schedule_time, self._components_height, self._x_axis_indent)
        self._axis.width = schedule_time

        # add axis and components
        self.addToGroup(self._axis)
        # self._axis.update_axis(schedule_time - 2, self._components_height, self._x_axis_indent)
        for component in self._components:
            self.addToGroup(component)
        # self.addToGroup(self._components)